#lang scheme

(provide (all-defined-out))

(require srfi/1
         redex/reduction-semantics
         (except-in syntax/parse id)
         (only-in "unstable/text.rkt" text<?))

(provide (rename-out [test-results test]))

;
; Basic
;

; id : a -> a
(define (id x)
  x)

; fix : ((a -> b) * a -> b) -> (a -> b)
(define ((fix f) a)
  (f (fix f) a))

; const : a -> b -> a
(define (const x)
  (lambda args x))

; truth : any -> bool
(define (truth x)
  (not (not x)))

; down-from : Nat -> List(Nat)
; (down-from n) = (n ... 0)
(define (down-from n)
  (if (zero? n)
    (list 0)
    (cons n (down-from (- n 1)))))

; up-to : Nat -> List(Nat)
; (up-to n) = (0 ... n)
(define up-to (compose reverse down-from))

; pair-of : (a -> Bool) * (b -> Bool) -> (a*b -> Bool)
(define ((pair-of p1 p2) val)
  (and (pair? val) (p1 (car val)) (p2 (cdr val))))

; ⟨⟩, tuple : (z → a) ... → z → (list a ...)
(define ((tuple . fs) z)
  (map (@ z) fs))
(define ⟨⟩ tuple)

; ×, cross : (a → a′) ... → (list a ...) → (list a′ ...)
(define ((cross . fs) as)
  (map (λ (f a) (f a)) fs as))
(define × cross)

; @ : a -> (a -> b) -> b
(define ((@ . args) f)
  (apply f args))

; Tests
(test-equal ((cross (curry apply +) (curry apply +) (curry apply +))
                    '((1 1 1)
                      (2 2 2)
                      (3 3 3)))
            '(3 6 9))

; pointwise : (b * ... -> c) * list((a -> b) * ...) -> a * ... -> c
(define ((pointwise f gs) . args)
  (f (map (apply @ args) gs)))

; memoize : ((a -> b) * a -> b) -> (a -> b)
(define (memoize g)
  (let ([h (make-hasheq)])
    (fix (λ (f x)
           (cond
             [(hash-ref h x #f) => id]
             [#t (let ((r (g f x)))
                   (hash-set! h x r)
                   r)])))))

; maybes->list : (list-of (maybe a)) -> (list-of a)
;   where (maybe a) = (a | #f)
(define (maybes->list as)
  (filter id as))

(define ≤    <=)
(define ≥    >=)
(define ≠    (compose not =))
(define all  andmap)
(define some ormap)
(define ∈    member)
(define ∘    compose)

; in-interval : ℝ (list ℝ ℝ) → 2
(define (in-interval y xz)
  (match-let ([(list x z) xz])
    (≤ x y z)))

;
; Lists
;

; concat : List(List(a)) -> List(a)
(define (concat ls)
  (apply append ls))

; ∑, sum : List(Num) -> Num
(define (sum l) (fold-list 0 + l))
(define ∑ sum)

; ∏, product : List(Num) -> Num
(define (product l) (fold-list 1 * l))
(define ∏ product)

; enumerate : List(a) -> List(Tuple(Nat, a))
(define (enumerate l)
  (fold-list-acc
    0
    (lambda (i) (+ 1 i))
    '()
    (lambda (i a tail) (cons (list i a) tail))
    l))

; repeat : Nat * a -> List(a)
(define (repeat n v)
  (fold-nat '() (curry cons v) n))

; inits : List(a) -> List(List(a))
(define (inits as)
  (map (lambda (n) (take as n)) (iota (+ 1 (length as)))))

; tails : List(a) -> List(List(a))
(define (tails as)
  (map (lambda (n) (drop as n)) (iota (+ 1 (length as)))))

; snoc : List(a) * a -> List(a)
(define (snoc as a)
  (append as (list a)))

; concat-map : (a -> List(b)) * List(a) -> List(b)
; (This is handy.)
(define (concat-map . args)
  (concat (apply map args)))

; unzip : list(a*...*z) -> list(a) * ... * list(z)
(define (unzip l)
  (apply zip l)) ; transpose

; ∪, union : list(a) ... -> list(a)
; Unlike lset-union, union preserves order.
(define (union-2 as bs) (append as (filter (λ (b) (not (member b as))) bs)))
(define (union . args)
  (fold-list '() union-2 args))
(define ∪ union)

; unlist : list(a) -> a
(define (unlist l)
  (match l
    [(list a) a]))

; subtract : list(a) list(a) -> list(a)
; Set difference
(define (subtract as bs)
  (fold-list as (λ (b r) (delete b r)) bs))

;
; Strings
;

; ->string : any -> string
(define (->string x)
  (format "~s" x))

; strip : string -> string
(define (strip s)
  ; TODO Why does #px"^\\s*|\\s*$" replace internal spaces?
  (regexp-replace #px"^\\s*" (regexp-replace #px"\\s*$" s "") ""))

;
; match
;

(define-syntax-rule (match-λ args ...)
  (match-lambda args ...))

(define-syntax-rule (match-λ* args ...)
  (match-lambda* args ...))

(define-syntaxes (lambda/match λ/match)
  (let ([t (syntax-rules ()
             [(_ (pat ...) body ...)
              (match-λ* [(list pat ...) body ...])])])
    (values t t)))

;
; Keywords
;

(define (make-keyword-procedure/dict f)
  (make-keyword-procedure
    (λ (kw-ids kw-args . rest)
      (apply f (make-immutable-hash (map cons kw-ids kw-args)) rest))))

(define (keyword-apply/dict f kws . rest)
  (let-values ([(kw-ids kw-args)
                (apply values
                       (unzip (sort (dict-map kws list) #:key car text<?)))])
    (keyword-apply f kw-ids kw-args (apply list* rest))))

(define-syntaxes (λ/kw-dict lambda/kw-dict)
  (let ([t (syntax-rules ()
             [(_ args body ...)
              (make-keyword-procedure/dict (lambda args body ...))])])
    (values t t)))

(define-syntax define/kw-dict
  (syntax-rules ()
    [(_ (f x ...) e ...)      (define/kw-dict f (λ/kw-dict (x ...) e ...))]
    [(_ (f x ... . xs) e ...) (define/kw-dict f (λ/kw-dict (x ... . xs) e ...))]
    [(_ f e ...)              (define f e ...)]))

; dict-extend : dict ... -> dict
(define (dict-extend . ds)
  ((apply compose
          (reverse (concat (map (curryr dict-map (curry curryr dict-set)) ds))))
   '()))

; dict-remove* : dict key ... -> dict
(define (dict-remove* d . ks)
  (fold-list d (λ (k d) (dict-remove d k)) ks))

;
; Primitive folds
;

; Nat = Zero | Succ Nat
; fold-nat : a * (a -> a) * Nat -> a
(define (fold-nat z s n)
  (cond
    ((= n 0) z)
    ((> n 0) (s (fold-nat z s (- n 1))))
    (else (error "Not a natural number:" n))))

; List = () | (Val . List)
; fold-list : b * (a * b -> b) * List(a) -> b
(define (fold-list c f l)
  (if (null? l)
    c
    (f (car l) (fold-list c f (cdr l)))))

; SnocList = () | (List . Val)
; fold-snoc-list : a * (b * a -> b) * List(a) -> b
(define (fold-snoc-list c f l)
  (if (null? l)
    c
    (f (fold-snoc-list c f (cdr l)) (car l))))

; TODO: How useful is this pattern? Should succ get more information?
; fold-list-acc : n * (n -> n) * a * (a * b -> b) * List(a) -> b
(define (fold-list-acc zero succ nil-c cons-f l)
  (let loop ((acc zero) (l l))
    (if (null? l)
      nil-c
      (cons-f acc (car l) (loop (succ acc) (cdr l))))))

; S-exp  = Sym | S-list
; S-list = () | (S-exp . S-list)
; fold-sexp : (Sym -> a) * a * (a * a -> a) * S-exp -> a
(define (fold-sexp sym-f nil-c cons-f sexp)
  (if (not (list? sexp))
    (sym-f sexp)
    (fold-slist sym-f nil-c cons-f sexp)))
; fold-slist : (Sym -> a) * a * (a * a -> a) * S-list -> a
(define (fold-slist sym-f nil-c cons-f slist)
  (if (null? slist)
    nil-c
    (cons-f (fold-sexp sym-f nil-c cons-f (car slist))
            (fold-slist sym-f nil-c cons-f (cdr slist)))))

;
; Syntax
;

(define-syntax-rule (expand-e e)
  (syntax->datum (expand #'e)))

(define-syntax-rule (expand-once-e e)
  (syntax->datum (expand-once #'e)))

;
; syntax/parse
;

(define-syntax-class (atom a)
  #:attributes ()
  #:description (->string a)
  (pattern x #:fail-unless (eq? (syntax-e #'x) a) #f))

;
; Test
;

; Tests
(test-equal (subtract '(1 2 3 4) '(2 3)) '(1 4))
(test-equal (union '(1 2) '(2 3) '(1 3)) '(1 2 3))
