#lang scheme

(require redex/reduction-semantics
         "grammar.rkt"
         "reduction.rkt"
         "typing.rkt"
         "compilation.rkt")

(define soundness?
  (let ([v? (redex-match SR v)])
    (λ (e)
      (let ([t (term (typeof () (reset ,e)))])
        (or (not t)
            (let ([es (apply-reduction-relation ->sr (term (reset ,e)))])
              (and (not (null? es))
                   (andmap (λ (e) (equal? t (term (typeof () ,e)))) es))))))))

(define-metafunction SR
  plt : e -> any
  [(plt (shift e)) (shift (plt e))]
  [(plt (reset e)) (reset (plt e))]
  [(plt (e_1 e_2)) ((plt e_1) (plt e_2))]
  [(plt (block (def f (x : (A ε)) : τ = e) r ...))
   (letrec ([f (λ (x) (plt e))]) (plt (block r ...)))]
  [(plt (block (val x : (A ε) = e) r ...))
   (let ([x (plt e)]) (plt (block r ...)))]
  [(plt (block e_0 r_1 ... e_i))
   (begin (plt e_0) (plt (block r_1 ... e_i)))]
  [(plt (block e)) (plt e)]
  [(plt x) x]
  [(plt number) number]
  [(plt string) string]
  [(plt inc) add1]
  [(plt dbl) (λ (x) (* x 2))]
  [(plt len) string-length])

(define (make-plt e)
  `(let-syntax ([shift (syntax-rules ()
                         [(_ e) (shift k (e k))])])
     ,(term (plt ,e))))

(define scheme/control-result
  (let ([ns (make-base-namespace)]
        [c (make-channel)])
    (parameterize ([current-namespace ns])
      (eval '(require scheme/control)))
    (λ (e)
      (parameterize ([current-namespace ns])
        (let ([t (thread (λ () (channel-put c (eval (make-plt e)))))])
          (cond [(sync/timeout 10 c) 
                 =>
                 (λ (v)
                   (if (procedure? v)
                       'procedure
                       v))]
                [else (begin (kill-thread t) 'timeout)]))))))

(define-metafunction SR
  [(result o) procedure]
  [(result (block (def f (x : τ_1) : τ_2 = e) f)) procedure]
  [(result e) e])

(define (reduction-result e r)
  (let recur ([e e] [t 10000])
    (if (zero? t)
        'timeout
        (match (apply-reduction-relation r e)
          [`() (term (result ,e))]
          [`(,e*) (recur e* (sub1 t))]
          [else (error 'reduction-result "~s reduces multiple ways" e)]))))

; Check that result has the expected type?
(define (evaluators-agree? e)
  (or (not (term (typeof () (reset ,e))))
      (equal? (reduction-result (term (reset ,e)) ->sr)
              (reduction-result (compile (term (reset ,e))) ->))
      (let ([r (scheme/control-result (term (reset ,e)))])
        (and (equal? r (reduction-result (term (reset ,e)) ->sr))
             (equal? r (reduction-result (compile (term (reset ,e))) ->))))))

(define (A-norm-preserves-type? e)
  (cond [(term (typeof () ,e))
         => (curry equal? (term (typeof () (ANF ,e))))]
        [else #t]))

(define (CPS-well-typed? e)
  (or (not (term (typeof () (reset ,e))))
      (term (typeof () (CPS (ANF (reset ,e)))))))

(define (main)
  (let ([per-round 10000]
        [width 10])
    (define-extended-language SR-test SR
      (x a b c d q))
    (let loop ([round 0])
      (let ([result (redex-check
                     SR-test e
                     (and (soundness? (term e))
                          (evaluators-agree? (term e))
                          (A-norm-preserves-type? (term e))
                          (CPS-well-typed? (term e)))
                     #:attempts per-round
                     #:print? #f)])
        (if (counterexample? result)
            result
            (begin
              (printf ".") (flush-output)
              (when (= (modulo round width) (sub1 width))
                (printf "\n"))
              (loop (add1 round))))))))

(provide main)
